@@ -70,6 +70,8 @@ from tensorflow.keras.layers.experimental.preprocessing import TextVectorization
 ############################################################################################
 import time
 import os
+import datetime
+
 from sklearn.metrics import balanced_accuracy_score, classification_report, confusion_matrix
 from sklearn.metrics import roc_auc_score
 from collections import defaultdict
@@ -96,7 +98,7 @@ from .preprocessing.preprocessing_text import preprocessing_text
 from .modeling.create_model import create_model
 from .models import basic, deep, big_deep, giant_deep, cnn1, cnn2
 from .modeling.train_model import train_model
-from .modeling.train_custom_model import train_custom_model, return_optimizer
+from .modeling.train_custom_model import train_custom_model
 from .modeling.predict_model import predict, predict_images, predict_text
 from .modeling.train_image_model import train_image_model
 from .modeling.train_text_model import train_text_model
@@ -104,7 +106,6 @@ from .modeling.train_text_model import train_text_model
 # Utils
 from .utilities.utilities import print_one_row_from_tf_dataset
 from .utilities.utilities import print_one_row_from_tf_label
-from .utilities.utilities import get_compiled_model
 from .utilities.utilities import check_if_GPU_exists, plot_history
 
 #############################################################################################
@@ -209,15 +210,38 @@ def fit(train_data_or_file, target, keras_model_type="basic", project_name="deep
             "image_height": default is "" (empty string). Needed only for "image" use case.
             "image_width": default is "" (empty string). Needed only for "image" use case.
             "image_channels": default is "" (empty string). Needed only for image use case. Number of channels.
+            'save_model_path': default is project_name/keras_model_type/datetime-hour-min/
+                        If you provide your own model path as a string, it will save it there.
     model_use_case: default is "" (empty string). If "pipeline", you will get back pipeline only, not model.
                 It is a placeholder for future purposes. At the moment, leave it as empty string.
     verbose = 1 will give you more charts and outputs. verbose 0 will run silently
                 with minimal outputs.
     """
     my_strategy = check_if_GPU_exists(1)
+    print() #### create a new line that's all ###
     model_options_copy = copy.deepcopy(model_options)
     keras_options_copy = copy.deepcopy(keras_options)
 
+    if isinstance(project_name,str):
+        if project_name == '':
+            project_name = "deep_autoviml"
+    else:
+        print('Project name must be a string and helps create a folder to store model.')
+        project_name = "deep_autoviml"
+
+    save_model_path = os.path.join(project_name,keras_model_type)
+    save_model_path = get_save_folder(save_model_path)
+    if not os.path.exists(save_model_path):
+        os.makedirs(save_model_path, exist_ok = True)
+        trials_saved_path = os.path.join(save_model_path, "trials")
+        os.makedirs(trials_saved_path, exist_ok = True)
+        save_artifacts_path = os.path.join(save_model_path, "artifacts")
+        os.makedirs(save_artifacts_path, exist_ok = True)
+        save_logs_path = os.path.join(save_model_path, "mylogs")
+        os.makedirs(save_logs_path, exist_ok = True)
+
+    print('Model and logs being saved in %s' %save_model_path)
+
     if keras_model_type.lower() in ['image', 'images', "image_classification"]:
         ###############   Now do special image processing here ###################################
         with my_strategy.scope():
@@ -242,8 +266,8 @@ def fit(train_data_or_file, target, keras_model_type="basic", project_name="deep
                 return
         ##########    E N D    O F    S T R A T E G Y    S C O P E   #############
         deep_model, cat_vocab_dict = train_image_model(deep_model, train_ds, valid_ds,
-                                            cat_vocab_dict, keras_options_copy,
-                                            project_name, save_model_flag)
+                                        cat_vocab_dict, keras_options_copy, model_options_copy,
+                                        project_name, save_model_flag)
         print(deep_model.summary())
         return deep_model, cat_vocab_dict
     elif keras_model_type.lower() in ['text', 'text classification', "text_classification"]:
@@ -268,8 +292,8 @@ def fit(train_data_or_file, target, keras_model_type="basic", project_name="deep
             else:
                 #### Use the text file given and split it into train and valid_ds ####
                 dft, model_options, full_ds, var_df, cat_vocab_dict, keras_options = load_train_data(
-                                    train_data_or_file, target, project_name, keras_options,
-                                    model_options, keras_model_type, verbose=verbose)
+                                    train_data_or_file, target, project_name, keras_options_copy,
+                                    model_options_copy, keras_model_type, verbose=verbose)
                 print('Loaded text classification file or dataframe using input given:')
                 ############## Split train into train and validation datasets here ###############
                 recover = lambda x,y: y
@@ -290,143 +314,147 @@ def fit(train_data_or_file, target, keras_model_type="basic", project_name="deep
         return deep_model, cat_vocab_dict
 
     ########    B E G I N N I N G    O F    S T R A T E G Y    S C O P E    #####################
-    #with my_strategy.scope():
+    with my_strategy.scope():
 
-    shuffle_flag = False
-    ####   K E R A S    O P T I O N S   - THESE CAN BE OVERRIDDEN by your input keras_options dictionary ####
-    keras_options_defaults = {}
-    keras_options_defaults["batchsize"] = ""
-    keras_options_defaults['activation'] = ''
-    keras_options_defaults['save_weights_only'] = True
-    keras_options_defaults['use_bias'] = True
-    keras_options_defaults["patience"] = "" ### patience of 20 seems ideal.
-    keras_options_defaults["epochs"] = "" ## 500 seems ideal for most scenarios ####
-    keras_options_defaults["steps_per_epoch"] = "" ### 10 seems ideal for most scenarios
-    keras_options_defaults['optimizer'] = ""
-    keras_options_defaults['kernel_initializer'] =  ''
-    keras_options_defaults['num_layers'] = ""
-    keras_options_defaults['loss'] = ""
-    keras_options_defaults['metrics'] = ""
-    keras_options_defaults['monitor'] = ""
-    keras_options_defaults['mode'] = ""
-    keras_options_defaults["lr_scheduler"] = ""
-    keras_options_defaults["early_stopping"] = True
-    keras_options_defaults["class_weight"] = {}
+        shuffle_flag = False
+        ####   K E R A S    O P T I O N S   - THESE CAN BE OVERRIDDEN by your input keras_options dictionary ####
+        keras_options_defaults = {}
+        keras_options_defaults["batchsize"] = ""
+        keras_options_defaults['activation'] = ''
+        keras_options_defaults['save_weights_only'] = True
+        keras_options_defaults['use_bias'] = True
+        keras_options_defaults["patience"] = "" ### patience of 20 seems ideal.
+        keras_options_defaults["epochs"] = "" ## 500 seems ideal for most scenarios ####
+        keras_options_defaults["steps_per_epoch"] = "" ### 10 seems ideal for most scenarios
+        keras_options_defaults['optimizer'] = "RMSprop"
+        keras_options_defaults['kernel_initializer'] =  ''
+        keras_options_defaults['num_layers'] = ""
+        keras_options_defaults['loss'] = ""
+        keras_options_defaults['metrics'] = ""
+        keras_options_defaults['monitor'] = ""
+        keras_options_defaults['mode'] = ""
+        keras_options_defaults["lr_scheduler"] = ""
+        keras_options_defaults["early_stopping"] = True
+        keras_options_defaults["class_weight"] = {}
 
-    list_of_keras_options = ["batchsize", "activation", "save_weights_only", "use_bias",
-                            "patience", "epochs", "steps_per_epoch", "optimizer",
-                            "kernel_initializer", "num_layers", "class_weight",
-                            "loss", "metrics", "monitor","mode", "lr_scheduler","early_stopping"]
+        list_of_keras_options = ["batchsize", "activation", "save_weights_only", "use_bias",
+                                "patience", "epochs", "steps_per_epoch", "optimizer",
+                                "kernel_initializer", "num_layers", "class_weight",
+                                "loss", "metrics", "monitor","mode", "lr_scheduler","early_stopping",
+                                "class_weight"]
 
-    keras_options = copy.deepcopy(keras_options_defaults)
-    if len(keras_options_copy) > 0:
-        print('Using following keras_options given as input:')
-        for key in list_of_keras_options:
-            if key in keras_options_copy.keys():
-                print('    %s : %s' %(key, keras_options_copy[key]))
-                keras_options[key] = keras_options_copy[key]
+        keras_options = copy.deepcopy(keras_options_defaults)
+        if len(keras_options_copy) > 0:
+            print('Using following keras_options given as input:')
+            for key in list_of_keras_options:
+                if key in keras_options_copy.keys():
+                    print('    %s : %s' %(key, keras_options_copy[key]))
+                    keras_options[key] = keras_options_copy[key]
 
-    list_of_model_options = ["idcols","modeltype","sep","cat_feat_cross_flag", "model_use_case",
-                            "nlp_char_limit", "variable_cat_limit", "csv_encoding", "header",
-                            "max_trials","tuner", "embedding_size", "tf_hub_model", "image_directory",
-                            'image_height', 'image_width', "image_channels"]
+        list_of_model_options = ["idcols","modeltype","sep","cat_feat_cross_flag", "model_use_case",
+                                "nlp_char_limit", "variable_cat_limit", "csv_encoding", "header",
+                                "max_trials","tuner", "embedding_size", "tf_hub_model", "image_directory",
+                                'image_height', 'image_width', "image_channels", "save_model_path"]
 
-    model_options_defaults = defaultdict(str)
-    model_options_defaults["idcols"] = []
-    model_options_defaults["modeltype"] = ''
-    model_options_defaults["save_model_format"] = ""
-    model_options_defaults["sep"] = ","
-    model_options_defaults["cat_feat_cross_flag"] = False
-    model_options_defaults["model_use_case"] = ''
-    model_options_defaults["nlp_char_limit"] = 30
-    model_options_defaults["variable_cat_limit"] = 30
-    model_options_defaults["csv_encoding"] = 'utf-8'
-    model_options_defaults["header"] = 0 ### this is the header row for pandas to read
-    model_options_defaults["max_trials"] = 30 ## number of Storm Tuner trials ###
-    model_options_defaults['tuner'] = 'storm'  ## Storm Tuner is the default tuner. Optuna is the other option.
-    model_options_defaults["embedding_size"] = "" ## this is the NLP embedding size minimum
-    model_options_defaults["tf_hub_model"] = "" ## If you want to use a pretrained Hub model, provide URL here.
-    model_options_defaults["image_directory"] = "" ## this is where images are input in form of folder
-    model_options_defaults['image_height'] = "" ## the height of the image must be given in number of pixels
-    model_options_defaults['image_width'] = "" ## the width of the image must be given in number of pixels
-    model_options_defaults["image_channels"] = "" ## number of channels in images provided
+        model_options_defaults = defaultdict(str)
+        model_options_defaults["idcols"] = []
+        model_options_defaults["modeltype"] = ''
+        model_options_defaults["save_model_format"] = ""
+        model_options_defaults["sep"] = ","
+        model_options_defaults["cat_feat_cross_flag"] = False
+        model_options_defaults["model_use_case"] = ''
+        model_options_defaults["nlp_char_limit"] = 30
+        model_options_defaults["variable_cat_limit"] = 30
+        model_options_defaults["csv_encoding"] = 'utf-8'
+        model_options_defaults["header"] = 0 ### this is the header row for pandas to read
+        model_options_defaults["max_trials"] = 30 ## number of Storm Tuner trials ###
+        model_options_defaults['tuner'] = 'storm'  ## Storm Tuner is the default tuner. Optuna is the other option.
+        model_options_defaults["embedding_size"] = "" ## this is the NLP embedding size minimum
+        model_options_defaults["tf_hub_model"] = "" ## If you want to use a pretrained Hub model, provide URL here.
+        model_options_defaults["image_directory"] = "" ## this is where images are input in form of folder
+        model_options_defaults['image_height'] = "" ## the height of the image must be given in number of pixels
+        model_options_defaults['image_width'] = "" ## the width of the image must be given in number of pixels
+        model_options_defaults["image_channels"] = "" ## number of channels in images provided
+        model_options_defaults['save_model_path'] = save_model_path
 
-    model_options = copy.deepcopy(model_options_defaults)
-    if len(model_options_copy) > 0:
-        print('Using following model_options given as input:')
-        for key in list_of_model_options:
-            if key in model_options_copy.keys():
-                print('    %s : %s' %(key, model_options_copy[key]))
-                model_options[key] = model_options_copy[key]
+        model_options = copy.deepcopy(model_options_defaults)
+        if len(model_options_copy) > 0:
+            print('Using following model_options given as input:')
+            for key in list_of_model_options:
+                if key in model_options_copy.keys():
+                    print('    %s : %s' %(key, model_options_copy[key]))
+                    model_options[key] = model_options_copy[key]
 
-    fast_models = ['deep_and_wide','deep_wide','wide_deep', 'wide_and_deep','deep wide',
-            'wide deep', 'fast','fast1', 'fast2', 'deep_and_cross', 'deep cross', 'deep and cross']
-    if keras_model_type.lower() in fast_models:
-        print('max_trials set to 10 for fast models. Please increase it if you want better performance...')
-        model_options["max_trials"] = 10
-    else:
-        if model_options["max_trials"] <= 20:
-            print('Max Trials : %s. Please increase max_trials if you want to better accuracy...' %model_options["max_trials"])
+        fast_models = ['deep_and_wide','deep_wide','wide_deep', 'wide_and_deep','deep wide',
+                'wide deep', 'fast','fast1', 'fast2', 'deep_and_cross', 'deep cross', 'deep and cross']
+        if keras_model_type.lower() in fast_models:
+            print('max_trials set to 10 for fast models. Please increase it if you want better performance...')
+            model_options["max_trials"] = 10
         else:
-            print('Max Trials : %s. Please lower max_trials if you want to run faster...' %model_options["max_trials"])
+            if model_options["max_trials"] <= 20:
+                print('Max Trials : %s. Please increase max_trials if you want to better accuracy...' %model_options["max_trials"])
+            else:
+                print('Max Trials : %s. Please lower max_trials if you want to run faster...' %model_options["max_trials"])
 
-    print("""
-#################################################################################
-###########     L O A D I N G    D A T A    I N T O   TF.DATA.DATASET H E R E  #
-#################################################################################
-        """)
-    dft, model_options, batched_data, var_df, cat_vocab_dict, keras_options = load_train_data(
-                                    train_data_or_file, target, project_name, keras_options,
-                                    model_options, keras_model_type, verbose=verbose)
+        print("""
+    #################################################################################
+    ###########     L O A D I N G    D A T A    I N T O   TF.DATA.DATASET H E R E  #
+    #################################################################################
+            """)
+        dft, model_options, batched_data, var_df, cat_vocab_dict, keras_options = load_train_data(
+                                        train_data_or_file, target, project_name, keras_options,
+                                        model_options, keras_model_type, verbose=verbose)
 
-    try:
-        data_size = cat_vocab_dict['DS_LEN']
-    except:
-        data_size = 10000
-        cat_vocab_dict['DS_LEN'] = data_size
+        try:
+            data_size = cat_vocab_dict['DS_LEN']
+        except:
+            data_size = 10000
+            cat_vocab_dict['DS_LEN'] = data_size
 
-    modeltype = model_options['modeltype']
+        modeltype = model_options['modeltype']
 
-    ##########  Perform keras preprocessing here by building all layers needed #############
-    print("""
-#################################################################################
-###########     K E R A S     F E A T U R E    P R E P R O C E S S I N G  #######
-#################################################################################
-        """)
+        ##########  Perform keras preprocessing here by building all layers needed #############
+        print("""
+    #################################################################################
+    ###########     K E R A S     F E A T U R E    P R E P R O C E S S I N G  #######
+    #################################################################################
+            """)
 
-    nlp_inputs, meta_inputs, meta_outputs, nlp_outputs = perform_preprocessing(batched_data, var_df,
-                                                cat_vocab_dict, keras_model_type,
-                                                keras_options, model_options,
-                                                verbose)
+        nlp_inputs, meta_inputs, meta_outputs, nlp_outputs = perform_preprocessing(batched_data, var_df,
+                                                    cat_vocab_dict, keras_model_type,
+                                                    keras_options, model_options,
+                                                    verbose)
 
-    if isinstance(model_use_case, str):
-        if model_use_case:
-            if model_use_case.lower() == 'pipeline':
-                ##########  Perform keras preprocessing only and return inputs + keras layers created ##
-                print('\nReturning a keras pipeline so you can create your own Functional model.')
-                return nlp_inputs, meta_inputs, meta_outputs, nlp_outputs
-            #### There may be other use cases for model_use_case in future hence leave this empty for now #
+        if isinstance(model_use_case, str):
+            if model_use_case:
+                if model_use_case.lower() == 'pipeline':
+                    ##########  Perform keras preprocessing only and return inputs + keras layers created ##
+                    print('\nReturning a keras pipeline so you can create your own Functional model.')
+                    return nlp_inputs, meta_inputs, meta_outputs, nlp_outputs
+                #### There may be other use cases for model_use_case in future hence leave this empty for now #
 
-    #### you must create a functional model here
-    print('\nCreating a new Functional model here...')
-    print('''
-#################################################################################
-###########     C R E A T I N G    A    K E R A S       M O D E L    ############
-#################################################################################
-        ''')
-    ######### this is where you get the model body either by yourself or sent as input ##
-    ##### This takes care of providing multi-output predictions! ######
-    model_body, keras_options =  create_model(use_my_model, nlp_inputs, meta_inputs, meta_outputs,
-                                        nlp_outputs, keras_options, var_df, keras_model_type,
-                                        model_options, cat_vocab_dict)
+        #### you must create a functional model here
+        print('\nCreating a new Functional model here...')
+        print('''
+    #################################################################################
+    ###########     C R E A T I N G    A    K E R A S       M O D E L    ############
+    #################################################################################
+            ''')
+        ######### this is where you get the model body either by yourself or sent as input ##
+        ##### This takes care of providing multi-output predictions! ######
+        model_body, keras_options =  create_model(use_my_model, nlp_inputs, meta_inputs, meta_outputs,
+                                            nlp_outputs, keras_options, var_df, keras_model_type,
+                                            model_options, cat_vocab_dict)
 
-    ###########    C O M P I L E    M O D E L    H E R E         #############
-    ### For auto models we will add input and output layers later. See below... #########
-    deep_model = model_body
+        ###########    C O M P I L E    M O D E L    H E R E         #############
+        ### For auto models we will add input and output layers later. See below... #########
+        deep_model = model_body
+
+    ##########    E N D    O F    S T R A T E G Y    S C O P E   #############
 
     if dft.shape[1] <= 100 :
         plot_filename = save_model_architecture(deep_model, project_name, keras_model_type, cat_vocab_dict,
-                         chart_name="model_before")
+                         model_options, chart_name="model_before")
         if plot_filename != "":
             display(Image(retina=True, filename=plot_filename))
     print("""
@@ -444,17 +472,18 @@ def fit(train_data_or_file, target, keras_model_type="basic", project_name="deep
         print('Training a %s model option...' %keras_model_type)
         deep_model, cat_vocab_dict = train_model(deep_model, batched_data, target, keras_model_type,
                         keras_options, model_options, var_df, cat_vocab_dict, project_name, save_model_flag, verbose)
-    ##########    E N D    O F    S T R A T E G Y    S C O P E   #############
+
     ##### Plot the model again after you have done model search #############
     if dft.shape[1] <= 100 :
         plot_filename = save_model_architecture(deep_model, project_name, keras_model_type, cat_vocab_dict,
-                        chart_name="model_after")
+                        model_options, chart_name="model_after")
         if plot_filename != "":
             display(Image(retina=True, filename=plot_filename))
     distributed_values = (deep_model, cat_vocab_dict)
     return distributed_values
 ############################################################################################
-def save_model_architecture(model, project_name, keras_model_type, cat_vocab_dict, chart_name="model_before"):
+def save_model_architecture(model, project_name, keras_model_type, cat_vocab_dict, 
+                    model_options, chart_name="model_before"):
     """
     This function saves the model architecture in a PNG file in the artifacts sub-folder of project_name folder
     """
@@ -464,23 +493,20 @@ def save_model_architecture(model, project_name, keras_model_type, cat_vocab_dic
     else:
         print('Project name must be a string and helps create a folder to store model.')
         project_name = "deep_autoviml"
-    save_model_path = os.path.join(project_name,keras_model_type)
-    if chart_name != 'model_before':
-        save_model_path = cat_vocab_dict['saved_model_path']
-    #plot_filename = project_name+'_'+keras_model_type+chart_name+'.png'
-    if not os.path.exists(save_model_path):
-        os.makedirs(save_model_path)
+    save_model_path = model_options['save_model_path']
     save_artifacts_path = os.path.join(save_model_path, "artifacts")
     try:
-        if not os.path.exists(save_artifacts_path):
-            os.makedirs(save_artifacts_path)
         plot_filename = os.path.join(save_artifacts_path,chart_name)+".png"
-        print('\nSaving model architecture after training in %s...will take time...' %plot_filename)
+        print('\nSaving model architecture...')
         tf.keras.utils.plot_model(model = model, to_file=plot_filename, dpi=72,
                         show_layer_names=True, rankdir="LR", show_shapes=True)
-        print('Model plot saved in file: %s' %plot_filename)
+        print('    model architecture saved in file: %s' %plot_filename)
     except:
-        print('Model plot not saved due to error. Continuing...')
+        print('Model architecture not saved due to error. Continuing...')
         plot_filename = ""
     return plot_filename
 #########################################################################################################
+def get_save_folder(save_dir):
+    run_id = time.strftime("model_%Y_%m_%d-%H_%M_%S")
+    return os.path.join(save_dir, run_id)
+######################################################################################
